{$F+}
{$M 16384,0,655360}
unit BScript;

interface

const
  HaltAtError: Boolean = True;
  CRLF: String[2] = #13 + #10;

type
  BigString = ^TBigStringRecord;
  TBigStringRecord = record
    StringSize: Word;
    Body: array [1..32000] of Char;
  end;

procedure NormalToBig(S: string; var Result: BigString);
function  BigToNormal(BS: BigString): String;
procedure Finalize(var BS: BigString);

procedure Interprete(InputFileName: String; var Result: BigString);

implementation

uses
  Streams, TestLib;

type
  TErrors = (erSentenceExpected, erSemicolonExpected, erExpressionExpected,
             erOperationExpected, erMissingOperande, erCloseBracketExpected,
             erOpenBracketExpected, erCommaExpected, erNumberExpected,
             erInvalideVariable, erIllegalID, erAssignExpected);

const
  Errors: array [TErrors] of String = (
    'Variable of ? operator expected.',
    '; expected.',
    'Expression expected.',
    '+ exprected.',
    'First oprande of + operation missing.',
    ') expected.',
    '( expected.',
    ', expected.',
    'Number expected.',
    'Variable value not define.',
    'Illegal identifier.',
    '= expected.'
  );


procedure NormalToBig;
begin
  GetMem(Result, 2 + Length(S));
  Result^.StringSize := Length(S);
  Move(S[1], Result^.Body, Result^.StringSize);
end;

function BigToNormal;
var
  RealLength: Integer;
  Result: String;
begin
  if BS^.StringSize > 255
  then RealLength := 255
  else RealLength := BS^.StringSize;
  {Copy Data}
  Result[0] := Char(RealLength);
  Move(BS^.Body, Result[1], RealLength);
end;

procedure Finalize;
begin
  FreeMem(BS, BS^.StringSize + 2);
  BS := Nil;
end;

(*****************************************************************************
   This procedure interpretate all program in BScript
   Grammar:
    Program ::= Sentence {; Sentence}
    Sentence ::= Assigment | ? Expression
    Expression ::= ' String ' | " String " | VarName | ( Expression ) |
                   Expression + Expression |
                   $( Expression , Number , Number)
    Assigment ::= VarName = Expression
    VarName ::= Letter
    String ::= Char {Char}
    Letter ::= 'A'..'Z', 'a'..'z'
*****************************************************************************)

function IntToStr(Value: Longint): String;
var
  Result: String;
begin
  Str(Value, Result);
  IntToStr := Result;
end;

procedure InternalError(const Message: String; Stream: PTextFileStream);
begin
  if HaltAtError
  then begin
    Writeln('PiBAS run-time error:');
    Writeln('  Syntax error: ', Message);
    Writeln('  Line #', Stream^.ErrorLine, ', column #', Stream^.ErrorSymbol,
            ' on "', Stream^.Lexeme, '"');
    Halt(200);
  end else begin
    QUIT(_PE, CRLF + 'PiBAS run-time error:' +
         CRLF +'  Syntax error: ' + Message +
         CRLF + '  Line #' + IntToStr(Stream^.ErrorLine) + ', column #' + IntToStr(Stream^.ErrorSymbol) +
         CRLF + '  on "' + Stream^.Lexeme + '"');
  end;
end;

procedure Error(ErrorCode: TErrors; const Stream: TTextFileStream);
begin
  if HaltAtError
  then begin
    Writeln('PiBAS run-time error:');
    Writeln('  Error code #', Ord(ErrorCode), '. Error: ', Errors[ErrorCode]);
    Writeln('  Line #', Stream.ErrorLine, ', column #', Stream.ErrorSymbol,
            ' on "', Stream.Lexeme, '"');
    Halt(100 + Ord(ErrorCode));
  end else begin
    QUIT(_WA, CRLF + 'PiBAS run-time error:' +
         CRLF + '  Error code #' + IntToStr(Ord(ErrorCode)) + '. Error: ' + Errors[ErrorCode] +
         CRLF + '  Line #' + IntToStr(Stream.ErrorLine) + ', column #' + IntToStr(Stream.ErrorSymbol) +
         CRLF + '  on "' + Stream.Lexeme + '"');
  end;
end;

function ConcatBS(var First, Second: BigString): BigString;
var
  Result: BigString;
begin
  {Crate Result}
  GetMem(Result, First^.StringSize + Second^.StringSize + 2);
  Result^.StringSize := First^.StringSize + Second^.StringSize;
  Move(First^.Body[1], Result^.Body[1], First^.StringSize);
  Move(Second^.Body[1], Result^.Body[1 + First^.StringSize], Second^.StringSize);
  ConcatBS := Result;
end;

function ConcatBSFree(var First, Second: BigString): BigString;
begin
  ConcatBSFree := ConcatBS(First, Second);
  Finalize(First);
  Finalize(Second);
end;

procedure Interprete;
var
  {Contain variables, if var not used - NIL}
  Vars: array ['A'..'Z'] of BigString;

  function ReadExpression(const Stream: TTextFileStream): BigString; forward;

  { This function read function defenintion from file and calulate it.
  }
  function CalcFunction(const Stream: TTextFileStream): BigString;
  var
    Result: BigString;
    Start, Len: Integer;
    Temp: BigString;
  begin
    if (Stream.GetNextLexeme <> symSymbol) or (Stream.ReadSymbol <> '(')
    then Error(erOpenBracketExpected, Stream);
    Result := ReadExpression(Stream);
    if (Stream.CurrentLexeme <> symSymbol) or (Stream.ReadSymbol <> ',')
    then Error(erCommaExpected, Stream);
    if Stream.GetNextLexeme <> symNumber
    then Error(erNumberExpected, Stream)
    else Start := Stream.ReadNumber;
    if (Stream.GetNextLexeme <> symSymbol) or (Stream.ReadSymbol <> ',')
    then Error(erCommaExpected, Stream);
    if Stream.GetNextLexeme <> symNumber
    then Error(erNumberExpected, Stream)
    else Len := Stream.ReadNumber;
    if (Stream.GetNextLexeme <> symSymbol) or (Stream.ReadSymbol <> ')')
    then Error(erCloseBracketExpected, Stream);
    {Check all parametrs}
    if Start <= 0
    then begin
      Len := Len + Start - 1;
      Start := 1;
    end;
    if Len > (Result^.StringSize - Start + 1)
    then Len := Result^.StringSize - Start + 1;
    if Len >= 1
    then begin
      GetMem(Temp, Len + 2);
      Move(Result^.Body[Start], Temp^.Body[1], Len);
      Temp^.StringSize := Len;
    end else begin
      GetMem(Temp, 2);
      Temp^.StringSize := 0;
    end;
    Finalize(Result);
    CalcFunction := Temp;
  end;

  { This function read Expression definition from stream
  }
  function ReadExpression(const Stream: TTextFileStream): BigString;
  var
    Operande: Boolean;
    Quit: Boolean;
    Result: BigString;
    Temp: BigString;
    ID: String;
  begin
    GetMem(Result, 2);
    Result^.StringSize := 0;
    Operande := True;   {Wait operande}
    Quit := False;
    repeat
      if Operande
      then begin
       case Stream.GetNextLexeme of
         symString: begin
           Operande := False;
           NormalToBig(Stream.ReadString, Temp);
           Result := ConcatBSFree(Result, Temp);
         end;{symString}
         symID: begin
           ID := Stream.ReadID;
           if (Length(ID) <> 1) or (not (UpCase(ID[1]) in ['A'..'Z']))
           then Error(erIllegalID, Stream)
           else ID[1] := UpCase(ID[1]);
           if Vars[ID[1]] = Nil
           then Error(erInvalideVariable, Stream);
           Operande := False;
           Temp := Result;
           Result := ConcatBS(Temp, Vars[ID[1]]);
           Finalize(Temp);
         end;
         symSymbol: begin
           case Stream.ReadSymbol of
             '+': Error(erMissingOperande, Stream);
             '(': begin
               Operande := False;
               Temp := ReadExpression(Stream);
               Result := ConcatBSFree(Result, Temp);
               if (Stream.CurrentLexeme <> symSymbol) or (Stream.ReadSymbol <> ')')
               then Error(erCloseBracketExpected, Stream);
             end; {')'}
             '$': begin
               Operande := False;
               Temp := CalcFunction(Stream);
               Result := ConcatBSFree(Result, Temp);
             end; {'$'}
             else Error(erExpressionExpected, Stream);
           end;
           Operande := False;
         end;
         else Error(erExpressionExpected, Stream);
       end; {case}
     end else begin
       if (Stream.GetNextLexeme <> symSymbol) or (Stream.ReadSymbol <> '+')
       then Quit := True
       else Operande := True;
     end;
    until Quit;
    ReadExpression := Result;
  end;

var
  FStream: ^TTextFileStream;
  TempString: BigString;
  ID: String;
  Index: Char;
begin
  {Initialization}
  GetMem(Result, 2);
  Result^.StringSize := 0;
  FillChar(Vars, SizeOf(Vars), 0);
  {Crate stream}
  New(FStream, Create(InputFileName));
  FStream^.ErrorHandler := InternalError;
  FStream^.GetNextLexeme;
  while FStream^.CurrentLexeme <> symEndStream do begin
    case FStream^.CurrentLexeme of
      symSymbol: begin
        {? opepator}
        if FStream^.ReadSymbol <> '?'
        then Error(erSentenceExpected, FStream^)
        else begin
          {Add to Result String}
          TempString := ReadExpression(FStream^);
          Result := ConcatBSFree(Result, TempString);
        end; {if ...}
      end; {symSymbol}
      symID: begin
        {Assignement}
        ID := FStream^.ReadID;
        if (Length(ID) <> 1) or (not (UpCase(ID[1]) in ['A'..'Z']))
        then Error(erIllegalID, FStream^)
        else ID[1] := UpCase(ID[1]);
        if (FStream^.GetNextLexeme <> symSymbol) or (FStream^.ReadSymbol <> '=')
        then Error(erAssignExpected, FStream^);
        if Vars[ID[1]] = Nil
        then Vars[ID[1]] := ReadExpression(FStream^)
        else begin
          TempString := ReadExpression(FStream^);
          Finalize(Vars[ID[1]]);
          Vars[ID[1]] := TempString;
        end;
      end; {symID}
      else Error(erSentenceExpected, FStream^);
    end; {case}
    {Check correct value of ';'}
    if FStream^.CurrentLexeme <> symEndStream
    then begin
     if (FStream^.CurrentLexeme <> symSymbol) or (FStream^.ReadSymbol <> ';')
     then Error(erSemicolonExpected, FStream^)
     else FStream^.GetNextLexeme;
    end; {if ..}
  end; {while ...}
  for Index := 'A' to 'Z' do
    if Vars[Index] <> Nil
    then Finalize(Vars[Index]);
  Dispose(FStream, Destroy);
end;

end.